////////////////////////////////////////////////////
// INI.CPP: implementation of the IniFile class
////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>
#include <mem.h>
#include "ini.h"

#define MAX_LENGTH 1024
#define MAX_SECTIONS 1024
// setting MAX_SECTIONS too high will cause Stack Overflows,
// at least when using MGL


//////////////////////////////////////////////////////////////////////
///===================================================================
///  Implementation of IniFile class
///===================================================================
//////////////////////////////////////////////////////////////////////



///////////////////////////////////
// IniFile::IniFile
///////////////////////////////////
//
// constructor for class IniFile
//
// Call is with the name of the file you want to open, and also whether
// you want the checking for section and item names to be case sensitive.
// Set case_sense to _TRUE_ if you want case sensitivity on.
//
IniFile::IniFile(char *filename)
{
	FILE *file;
	char line[MAX_LENGTH];
	int length, i, j;

	
	number_of_lines = 0;
	number_of_sections = 0;

	file = fopen(filename, "rb");
	if (file == NULL)
	{
		ini_file_read = _FALSE_;
		return;
	}

	//
	// if we have opened the ini file, I am assuming that what comes below
	// will work (ie that the file actually is an ini file and not something
	// else
	ini_file_read = _TRUE_;

	fseek(file, 0, SEEK_SET);

	//
	// count the number of lines in the open file
	// this does mean we read the file twice, but it doesn't add
	// too much overhead, and makes things that much more flexible
	while (!feof(file))
	{
		fgets(line, MAX_LENGTH, file);
		++number_of_lines;
	}

	inifile = new char*[number_of_lines];

	//
	// read in the open file into the inifile variable
	fseek(file, 0, SEEK_SET);
	for (i=0; i<number_of_lines; i++)
	{
		memset(line, 0x00, MAX_LENGTH);
		fgets(line, MAX_LENGTH, file);
		length = strlen(line);
		++length;
		inifile[i] = new char[length];
		memset(inifile[i], 0x00, length);
		strcpy(inifile[i], line);
	}

	CountNumberOfSections();

	//
	// if we didn't find any sections, delete the file from memory and
	// set ini_file_read to _FALSE_ so that none of the public calls will
	// do anything
	//
	if (number_of_sections == 0)
	{
		ini_file_read = _FALSE_;
		//delete the various lines
		for (int i=0; i<number_of_lines; i++)
		{
			delete [] inifile[i];
		}

		// delete the array of pointers
		delete [] inifile;

		return;
	}
	
	section_names = new char*[number_of_sections];

	GetSectionNames();

	entries = new IniEntry*[number_of_sections];

	// need a special case for the last section, as it can't be passed the
	// line position for the next section start position. In this special
	// case, it gets passed the total number of lines in the ini file.
	for (i=0; i<number_of_sections-1; i++)
	{
		entries[i] = new IniEntry(inifile, section_pos[i], section_pos[i+1]);
	}
	entries[number_of_sections-1] = new IniEntry(inifile, section_pos[number_of_sections-1], number_of_lines);

	fclose(file);
}



///////////////////////////////////////////
// IniFile::~IniFile()
///////////////////////////////////////////
//
// destructor for class IniFile
//
IniFile::~IniFile()
{
	if (ini_file_read)
	{
		//delete the various lines
		for (int i=0; i<number_of_lines; i++)
		{
			delete [] inifile[i];
		}

		// delete the array of pointers
		delete [] inifile;
	}
}



////////////////////////////////////////////////
//
// This skips spaces and TABS until it comes to the first letter. If
// this character is a '[', then it will return true, otherwise it will
// return _FALSE_.
//
////////////////////////////////////////////////
// FirstLetterIsOpenBracket(line)
////////////////////////////////////////////////
char IniFile::FirstLetterIsOpenBracket(int line)
{
	int length = strlen(inifile[line]);

	open_bracket_pos = 0;
	
	for (int i=0; i<length; i++)
	{
		switch(inifile[line][i])
		{
			case '[':
				open_bracket_pos = i;
				return _TRUE_;
				break;
			case ' ':
				break;
			case '\t':	// TAB
				break;
			default:
				return _FALSE_;
		}
	}
	return _FALSE_;
}



///////////////////////////////////////////
//
// This searches for a ']' occuring somewhere in the line of characters.
// If it encounters a ';', then it returns immediately with a false
// value.
///////////////////////////////////////////
// ContainsCloseBracket(line)
///////////////////////////////////////////
char IniFile::ContainsCloseBracket(int line)
{
	int length = strlen(inifile[line]);

	close_bracket_pos = 0;

	for (int i=open_bracket_pos+1; i<length; i++)
	{
		switch(inifile[line][i])
		{
			case ']':
				close_bracket_pos = i;
				return _TRUE_;
				break;
			case ';':
			case '\r':
			case '\n':
				return _FALSE_;
				break;
			default:
				break;
		}
	}
	return _FALSE_;
}



/////////////////////////////////////
//
// This must only be called in the constructor. It goes through the file
// that is now in inifile, and determines which lines of it contain section
// headers. A section header is defined as a line that has both a [ and ]
// that do not have a ; between them.
// It initialises section_pos with the number of sections that it finds,
// and places the line number of the sections into section_pos.
/////////////////////////////////////
// CountNumberOfSections()
/////////////////////////////////////
void IniFile::CountNumberOfSections()
{
	long int section_lines[MAX_SECTIONS];
	
	for (int i=0; i<number_of_lines; i++)
	{
		if (FirstLetterIsOpenBracket(i))
		{
			if(ContainsCloseBracket(i))
			{
				section_lines[number_of_sections] = i;
				++number_of_sections;
			}
		}
	}
	
	section_pos = new long int[number_of_sections];
	memcpy(section_pos, section_lines, number_of_sections*sizeof(long));
	
}



////////////////////////////////
//
// This is used to get the names of the various sections. It does this
// by grabbing the text between the opening [ and the closing ] and
// copying it into the appropriate section_names entry.
// It uses FirstLetterIsOpen and ContainsCloseBracket simply because
// those two functions set open_bracket_pos and close_bracket_pos
// respectively, which this uses to determine which part of the line to
// copy.
// Section headers do not have the brackets copied.
///////////////////////////////
// GetSectionNames()
///////////////////////////////
void IniFile::GetSectionNames()
{
	int length;
	int i, j;
	
	for (i=0; i<number_of_sections; i++)
	{
		FirstLetterIsOpenBracket(section_pos[i]);
		ContainsCloseBracket(section_pos[i]);
		length = close_bracket_pos - open_bracket_pos;
		section_names[i] = new char[length];
		memset(section_names[i], 0x00, length);
		char *p = inifile[section_pos[i]];
		p = p + open_bracket_pos + 1;
		for (j=0; j<length-1; j++)
		{
			strncpy(section_names[i], p, length-1);
		}
	}
}



///////////////////////////////////////
//
// This returns the number of the section entry that contains
// the section name specified. If the section name is not found,
// it returns -1.
//
///////////////////////////////////////
// FindSection(char *section)
///////////////////////////////////////
int IniFile::FindSection(char *section)
{
	int i;

	for (i=0; i<number_of_sections; i++)
	{
		if (!_stricmp(section, section_names[i]))
			return i;
	}

	return -1;
}


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
////  IniEntry code
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////
// IniEntry initialise function
///////////////////////////////////////////////////////////////////
//
// Pass it the file that has been read from memory, and also the
// start line of the section.
//
// This code does NOT scan over the section header - this has been
// tested. Nor does it scan over the next section header.
//
// A valid entry must have at least 1 character before an = sign,
// and also a ; must not appear before either this first character, or
// the equals sign.
//
IniEntry::IniEntry(char **fileimage, int first_line, int last_line)
{
	int i, number_of_lines;
	int j;

	number_of_entries = 0;
	first_non_space = -1;
	equals_position = -1;
	last_letter = -1;

	number_of_lines = last_line - first_line;

	// first we need to count the number of valid entries in this
	// section
	for (i=0; i<number_of_lines; i++)
	{
		if (HasEntryName(fileimage[first_line+i]))
		{
			if (ContainsEqualsSign(fileimage[first_line+i]))
			{
				++number_of_entries;
			}
		}
	}

	entry_names = new char*[number_of_entries];
	entry_values = new char*[number_of_entries];

	// j is used to keep track of where we are in entry_names and entry_values
	// can't use i because not all lines in number_of_lines will contain
	// an entry
	j = 0;
	for (i=0; i<number_of_lines; i++)
	{
		if (HasEntryName(fileimage[first_line+i]))
		{
			if(ContainsEqualsSign(fileimage[first_line+i]))
			{
				AddEntryName(fileimage[first_line+i], j);
				AddEntryValue(fileimage[first_line+i], j);
				++j;
			}
		}
	}
	
}



/////////////////////
//
// Destructor for IniEntry
//
/////////////////////
// ~IniEntry()
/////////////////////
IniEntry::~IniEntry()
{
	int i;
	
	for (i=0; i<number_of_entries; i++)
	{
		delete [] entry_names[i];
		delete [] entry_values[i];
	}
	delete [] entry_names;
	delete [] entry_values;
}


///////////////////////////////////////////
//
// This scans a line of text from the beginning looking for a non-whitespace
// character. If it finds a non-whitespace character, it returns _TRUE_ and
// sets first_non_space, unless that character was an = or a ;
//
// If it can't find a non-whitespace character, it returns _FALSE_.
///////////////////////////////////////////
// HasEntryName(char *textline)
///////////////////////////////////////////
char IniEntry::HasEntryName(char *textline)
{
	int length = strlen(textline);

	first_non_space = -1;
	for (int i=0; i<length; i++)
	{
		switch(textline[i])
		{
			case ' ':
			case '\t':	//TAB
				break;

			// exit on an '=', ';', '\r', '\n'
			case '=':
			case ';':
			case '\r':
			case '\n':
				return _FALSE_;

			default:
				first_non_space = i;
				return _TRUE_;
		}
	}

	return _FALSE_;
}


/////////////////////////////////////////////////
//
// This function scans a line of text to see if an = sign appears
// in it. It must always be called after HasEntryName has been called,
// as it uses first_non_space.
//
// Upon success, it sets equals_position to the relevant position in
// the line of text, and returns _TRUE_.
/////////////////////////////////////////////////
// ContainsEqualsSign(char *textline)
/////////////////////////////////////////////////
char IniEntry::ContainsEqualsSign(char *textline)
{
	int length = strlen(textline);
	int i, j;

	j = first_non_space+1;

	equals_position = -1;

	for (i=j; i<length; i++)
	{
		switch(textline[i])
		{
			case '=':
				equals_position = i;
				return _TRUE_;

			case '\r':
			case '\n':
			case ';':
				return _FALSE_;

			default:
				break;
		}
	}

	return _FALSE_;
}


/////////////////////////////////////////////////////
//
// Grab the position of the last character of the entry name
// before the equals sign
/////////////////////////////////////////////////////
// LastLetterBeforeEquals(char *textline)
/////////////////////////////////////////////////////
void IniEntry::LastLetterBeforeEquals(char *textline)
{
	int i;

	last_letter = -1;
	// need to go to first_non_space-1 to catch single letter entry names
	for (i=equals_position; i>first_non_space-1; i--)
	{
		switch(textline[i])
		{
			case '=':
			case ' ':
			case '\t':
				break;

			default:
				last_letter = i;
				return;
				break;
		}
	}
}


/////////////////////////////////////////////////////////////
//
// This determines an entry name. It adds 1 character onto the
// length of the entry name to put a NULL character at the end of
// the string, to stop rubbish being printed.
/////////////////////////////////////////////////////////////
// AddEntryName(char *textline, int entry_name)
/////////////////////////////////////////////////////////////
void IniEntry::AddEntryName(char *textline, int entry_number)
{
	int length;

	LastLetterBeforeEquals(textline);
	if (last_letter == -1)
		return;

	length = last_letter - first_non_space + 1;

	entry_names[entry_number] = new char[length+1];
	memset(entry_names[entry_number], 0x00, length+1);
	char *p = textline;
	p += first_non_space;
	strncpy(entry_names[entry_number], p, length);

}



//////////////////////////////////////////////////////////////
//
// This is called to add a value to entry_values from a specific
// line of text. Only call this from the construction function.
// It requires that the location of the equals sign has been identified.
// The entry_number is the entry_value to enter
//////////////////////////////////////////////////////////////
// AddEntryValue(char *textline, int entry_number)
//////////////////////////////////////////////////////////////
void IniEntry::AddEntryValue(char *textline, int entry_number)
{
	int length;

	if (FirstLetterAfterEquals(textline))
	{
		LastLetterAfterEquals(textline);
		RemoveTrailingSpaces(textline);

		length = last_after_equals - first_after_equals + 1;

		entry_values[entry_number] = new char[length+1];
		memset(entry_values[entry_number], 0x00, length+1);
		char *p = textline;
		p += first_after_equals;
		strncpy(entry_values[entry_number], p, length);
	}
	else
	{
		entry_values[entry_number] = new char[1];
		memset(entry_values[entry_number], 0x00, 1);
	}
}
	
	

/////////////////////////////////////////////////////
//
// This scans for the first non-whitespace letter after the
// equals sign. If it encounters a ';', it stops the scan.
//
/////////////////////////////////////////////////////
// FirstLetterAfterEquals(char *textline)
/////////////////////////////////////////////////////
char IniEntry::FirstLetterAfterEquals(char *textline)
{
	int i;
	int length = strlen(textline);

	first_after_equals = -1;
	for (i=equals_position+1; i<length; i++)
	{
		switch(textline[i])
		{
			case ' ':
			case '\t':	// TAB
				break;

			// exit on ';', '\r', '\n'
			case ';':
			case '\r':
			case '\n':
				return _FALSE_;

			default:
				first_after_equals = i;
				return _TRUE_;
		}
	}

	return _FALSE_;
}


////////////////////////////////////////////////////
//
// This finds the last letter of the value. It scans until
// it finds a \r, \n, or a ;.
//
////////////////////////////////////////////////////
// LastLetterAfterEquals(char *textline)
////////////////////////////////////////////////////
void IniEntry::LastLetterAfterEquals(char *textline)
{
	int i;
	int length = strlen(textline);

	last_after_equals = length;
	
	for (i=first_after_equals+1; i<length; i++)
	{
		switch(textline[i])
		{
			case ';':
			case '\r':
			case '\n':
				last_after_equals = i;
				return;
				break;

			default:
				break;
		}
	}
}



///////////////////////////////////////////////////
//
// This is called after LastLetterAfterEquals to remove
// and white-space at the end of the value. This is primarily
// used to get rid of comments, but also lets us determine the
// exact length of the value string.
//
///////////////////////////////////////////////////
// RemoveTrailingSpaces(char *textline)
///////////////////////////////////////////////////
void IniEntry::RemoveTrailingSpaces(char *textline)
{
	int i, j;

	j = last_after_equals-1;

	for (i=j; i>first_after_equals-1; i--)
	{
		switch(textline[i])
		{
			case ' ':
			case '\t':
				break;

			default:
				last_after_equals = i;
				return;
				break;
		}
	}
}
				

////////////////////////////////////
//
// This returns the number of the entry that contains
// the entry name specified. If the entry name is not found,
// it returns -1.
//
////////////////////////////////////
// FindEntry(char *entry)
////////////////////////////////////
int IniEntry::FindEntry(char *entry)
{
	int i;

	for (i=0; i<number_of_entries; i++)
	{
		if (!stricmp(entry, entry_names[i]))
			return i;
	}

	return -1;
}






/////////////////////////////////////////////////////////////////
/////   P U B L I C     F U N C T I O N S
/////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////
//
// This should be the function used to retrieve information
// from an ini file. Call this passing the section name and the
// name of the value you wish to retrieve from that section.
//
// It returns NULL if either the section name or the entry name
// are invalid.
//
// It is case insensitive.
//
///////////////////////////////////////////////////
// GetValue(char *section, char *entry)
///////////////////////////////////////////////////
char* IniFile::GetValue(char *section, char *entry)
{
	int s_number;

	s_number = FindSection(section);

	if (s_number == -1)
		return NULL;

	return entries[s_number]->EntryValue(entry);
}

///////////////////////////////////////
//
// This should only be called from within the GetValue function.
//
// It returns the string that makes up the value for the
// specified entry.
//
///////////////////////////////////////
// EntryValue(char *entry)
///////////////////////////////////////
char* IniEntry::EntryValue(char *entry)
{
	int entry_number;

	entry_number = FindEntry(entry);

	// if there wasn't a valid entry name, return null.
	if (entry_number == -1)
		return NULL;

	// otherwise, there was a valid entry name, so return
	// the value
	return entry_values[entry_number];
}


/////////////////////////////////////////////////
//
// 4 very short functions that simply return values that
// might be useful for when you are creating a list of what
// an ini file contains.
//
/////////////////////////////////////////////////

// SectionNames();
char** IniFile::SectionNames()
{
	return section_names;
}

// NumberOfSections();
int IniFile::NumberOfSections()
{
	return number_of_sections;
}

// SectionEntries(char *section)
char** IniFile::SectionEntries(char *section)
{
	int s_number = FindSection(section);
	if (s_number == -1)
		return NULL;
	return entries[s_number]->EntryNames();
}

// NumberOfEntries(char *section)
int IniFile::NumberOfEntries(char *section)
{
		int s_number = FindSection(section);
	if (s_number == -1)
		return NULL;
	return entries[s_number]->NumberOfEntries();
}

char** IniEntry::EntryNames()
{
	return entry_names;
}

int IniEntry::NumberOfEntries()
{
	return number_of_entries;
}
